=begin
#-------------------------------------------------------------------------------------------------------------------------------------------------
#*************************************************************************************************
# Designed April / June 2008 by Fredo6

# Permission to use, copy, modify, and distribute this software for 
# any purpose and without fee is hereby granted, provided that the above
# copyright notice appear in all copies.

# THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#-----------------------------------------------------------------------------
# Name			:   PolygonOnSurface.rb
# Original Date	:   04 June 2008 - version 1.2
# Type			:   Sketchup Tools
# Description	:   Polygon and Circles on Surface (part of the suite of Tools to draw on a surface)
# Menu Items	:   Tools --> "Polygon on Surface", "Circle on Surface"
# Toolbar		:   Name = "Surface Operations"
# Context Menu	:   All options
# Usage			:   See Tutorial and Quick Ref Card in PDF format
#-------------------------------------------------------------------------------------------------------------------------------------------------
#*************************************************************************************************
=end

require 'sketchup.rb'
require 'LibTraductor.rb'			# for language translation
require 'LibOnSurface.rb'		# for basic routines

module SUToolsOnSurface

#Constants for PolygonOnSurface Module (do not translate)	
DOS_ICON_POLYGON_LINE = "Polygon_Line"
DOS_ICON_POLYGON_CLINE = "Polygon_Cline"
DOS_CURSOR_POLYGON_LINE = "Polygon_Line"
DOS_CURSOR_POLYGON_CLINE = "Polygon_Cline"
DOS_ICON_CIRCLE_LINE = "Circle_Line"
DOS_ICON_CIRCLE_CLINE = "Circle_Cline"
DOS_CURSOR_CIRCLE_LINE = "Circle_Line"
DOS_CURSOR_CIRCLE_CLINE = "Circle_Cline"


STR_Polygon_Title = ["Polygon on Surface",
	              "|FR| Polygone sur une surface"]
STR_Circle_Title = ["Circle on Surface",
	              "|FR| Cercle sur une surface"]
MMSG_Polygon_Origin = ["Pick Initial Point",
	               "|FR| Cliquer Point Initial"]
MSG_Polygon_End = ["Pick End Point",
	            "|FR| Cliquer Point final"]
MSG_Polygon_Execution = ["Polygon On Surface",
	                  "|FR| Polygone sur une surface"]
				 
#--------------------------------------------------------------------------------------------------------------
# Top Calling functions: create the classes and launch the tools
#--------------------------------------------------------------------------------------------------------------			 				   

def SUToolsOnSurface.launch_polygon
	@tool_polygon = TOSToolPolygon.new false unless @tool_polygon
	Sketchup.active_model.select_tool @tool_polygon
end

def SUToolsOnSurface.launch_circle
	@tool_circle = TOSToolPolygon.new true unless @tool_circle
	Sketchup.active_model.select_tool @tool_circle
end

#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# TOSToolPolygon: Tool to mimic Sketchup Polygon Mesure on surface
#~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
					
class TOSToolPolygon

def initialize(circle)
	@circle = circle
	@linemode = true

	#Loading strings and cursors
	Traductor.load_translation SUToolsOnSurface, /MSG_/, binding, "@msg_"
	if (circle)
		@idcursor_polygon_line = OFSG.create_cursor DOS_CURSOR_CIRCLE_LINE, 3, 31
		@idcursor_polygon_cline = OFSG.create_cursor DOS_CURSOR_CIRCLE_CLINE, 3, 31
		@nb_seg = OFS_DEFAULT_CircleSegments
	else
		@idcursor_polygon_line = OFSG.create_cursor DOS_CURSOR_POLYGON_LINE, 3, 31
		@idcursor_polygon_cline = OFSG.create_cursor DOS_CURSOR_POLYGON_CLINE, 3, 31
		@nb_seg = OFS_DEFAULT_PolygonSegments
	end
	
	#initializing variables
	@ip_origin = Sketchup::InputPoint.new
	@ip_end = Sketchup::InputPoint.new
	@ip = Sketchup::InputPoint.new
	@prev_dist = 0
	@prev_dist0 = 0
	@prev_vec = nil
	@prev_vec0 = nil
	
	#Initializing parameters
	@option_group = OFS_DEFAULT_Group
	@option_cpoint = OFS_DEFAULT_CPoint_L
	@option_nocurves = ! OFS_DEFAULT_GenCurve
	@option_diameter = OFS_DEFAULT_Diameter
	
	#Group handling
	@ofsgrp = CommonGroup.create
	
end

def activate
	@model = Sketchup.active_model
	@selection = @model.selection
	@entities = @model.active_entities
			
	@option_group = @ofsgrp.get_option_group		
	@pt_origin = nil
	@pt_end = nil
	@state = STATE_ORIGIN
	@enter_down = false
	@pts_radius = []
	@distance = 0
	@parcours_radius = []
	@pts_vertices = []
	@pts_contour = []
	@edges = nil
	@moved = false
	
	info_show
	
	#OFSG.watchlist_clear
	#OFSG.watchlist_clear_face
end

def deactivate(view)
	view.invalidate
end

def execute_to_distance(mark, vecdir, len)
	parcours = Junction.to_distance mark, vecdir, len
	marklast = parcours.last
	@mark_end = marklast
	@pt_end = marklast.pt
	@face_end = marklast.face

	#Drawing the parcours
	execute_shape	
end

def compute_path_radius(store=false)
	return @pts_radius = [] unless @pt_origin && @pt_end

	parcours = Junction.calculate @mark_origin, @mark_end
	@pts_radius = []
	parcours.each { |mk| @pts_radius.push mk.pt }
	compute_distance
	if store
		@parcours_radius = parcours	#Storage for re-use after operation
	end
	return @pts_radius unless parcours.length >= 2
	compute_center_direction(parcours)
	return @pts_radius unless @vecref.valid?
	compute_polygon_contour(@mkcenter, @vecref, @distance)
	
	return @pts_radius
end
	
#Computing the parameters of the polygon	
def compute_center_direction(parcours)
	#By Radius
	unless @option_diameter
		@mkcenter = parcours[0]
		@radius = @distance
		@vecref = parcours[0].pt.vector_to parcours[1].pt
		return
	end	
	
	#by Diameter
	d = 0.0
	@radius = @distance * 0.5
	nb = parcours.length - 2
	for i in 0..nb
		mk1 = parcours[i]
		mk2 = parcours[i+1]
		d += mk2.pt.distance mk1.pt
		next if d < @radius
		if d == @radius
			@mkcenter = mk2
			@vecref = mk2.pt.vector_to mk1.pt
			return
		end
		@vecref = mk2.pt.vector_to mk1.pt
		pt = mk2.pt.offset @vecref, d - @radius
		@mkcenter = OFSG.mark(pt, mk2.face, nil, nil, nil)
		break
	end	
end
		
def compute_polygon_contour(mark, vecdir, distance)
	mark = @mkcenter
	vecdir = @vecref
	distance = @distance
	origin = mark.pt
	angle_unit = 2 * Math::PI / @nb_seg
	normal = (mark.face) ? mark.face.normal : Z_AXIS
	lst_vert = []
	nb = @nb_seg - 1
	for i in 0..nb
		t = Geom::Transformation.rotation origin, normal, angle_unit * i
		vec = vecdir.transform t
		parcours = Junction.to_distance mark, vec, @radius
		lst_vert.push parcours.last
	end
	lst_vert.push lst_vert.first
	lst_vert.first.signature = true
	
	@lmk_contour = [lst_vert.first]
	nb = lst_vert.length - 2
	for i in 0..nb
		parcours = Junction.calculate lst_vert[i], lst_vert[i+1]
		parcours[1..-1].each { |mk| @lmk_contour.push mk } if parcours.length > 1
		@lmk_contour.last.signature = true
	end
	
	#computing the points in the polygon contour
	@pts_contour = []
	@lmk_contour.each { |mk| @pts_contour.push mk.pt }
	@list_coseg = []
	OFSG.compute_coseg(@lmk_contour, @list_coseg) if @linemode
	@pts_vertices = []
	lst_vert.each { |mk| @pts_vertices.push mk.pt }
	
end
	
#Computing distance
def compute_distance
	return 0 if @pts_radius.length < 2
	nb = @pts_radius.length - 2
	@distance = 0.0.to_l
	for i in 0..nb
		@distance += @pts_radius[i].distance @pts_radius[i+1]
	end
end
	
def execute_drawing
	
	compute_path_radius true
	
	return if @pts_contour.length < 2
	
	#Storing distance for further Redo
	push_previous @pts_radius

	@model.start_operation @msg_Polygon_Execution
	
	#identifying the Group if needed
	if @option_group
		grp = @ofsgrp.get_current
		unless grp
			@model.abort_operation
			return
		end	
		entities = grp.entities
	else
		entities = @entities
	end
	
	if @linemode
		@edges = []
		attr = "Polygon L ---" + Time.now.to_i.to_s
		OFSG.commit_line entities, @pts_contour, attr, @option_nocurves, @list_coseg, @edges, @pts_vertices
	else
		nb = @pts_contour.length - 2
		attr = "Polygon C ---" + Time.now.to_i.to_s
		for i in 0..nb
			cline = entities.add_cline @pts_contour[i], @pts_contour[i+1]
			OFSG.set_cline_attribute cline, attr
		end
	end	
	cpoint = entities.add_cpoint @mkcenter.pt		#center of the polygon
	OFSG.set_cline_attribute cpoint, attr
	if @option_cpoint
		@pts_contour.each do |pt| 
			cpoint = entities.add_cpoint pt 
			OFSG.set_cline_attribute cpoint, attr
		end
	end	
	@model.commit_operation
end

#Top routine to calculate polygon shape		
def execute_shape
	execute_drawing
	set_state STATE_ORIGIN
end
	
def onCancel(flag, view)
	#User did an Undo
	case flag
	when 1, 2	#Undo or reselect the tool
		activate
		return
	when 0	#user pressed Escape
		return  if (@state == STATE_ORIGIN)
		set_state @state - 1
	end
end

#setting the right cursor
def onSetCursor
	UI::set_cursor (@linemode) ? @idcursor_polygon_line : @idcursor_polygon_cline
end

#Control the states of the tool
def set_state(state)
	@state = state
	if state == STATE_EXECUTION 
		execute_shape
	elsif state <= STATE_END
		@distance = 0.0
		@pts_vertices = []
		@pts_contour = []
	end
	info_show
end

#Contextual menu
def getMenu(menu)
	if (@state >= STATE_END)
		menu.add_item(@msg_MnuDone) { execute_shape }
		menu.add_separator
	end	
	if (@prev_dist > 0)
		tx = Traductor[MSG_MnuRedo, @prev_dist.to_l]
		menu.add_item(tx) { menu_redo }
		menu.add_separator
	end
	option_contextual_menu menu
	true
end

#Populate the options in the Contextual menu
def option_contextual_menu(menu)	
	txcur = @msg_MnuCurrent
	text = @msg_MnuGroup + " #{txcur} " + Traductor[DLG_EnumYesNo[(@option_group) ? 'Y' : 'N']] + 
	       ") --> " + OFS___Group
	menu.add_item(text) { self.change_option_group }
	text = @msg_MnuCPoint + " #{txcur} " + Traductor[DLG_EnumYesNo[(@option_cpoint) ? 'Y' : 'N']] + 
	       ") --> " + OFS___CPoint
	menu.add_item(text) { self.change_option_cpoint }
	text = @msg_MnuDiameter + " #{txcur} " + Traductor[(@option_diameter) ? MSG_Diameter : MSG_Radius] + 
	       ") --> " + OFS___Diameter
	menu.add_item(text) { self.change_option_diameter }
end

def change_option_linemode
	@linemode = !@linemode
	onSetCursor
end

def change_option_group
	@option_group = @ofsgrp.set_option_group(!@option_group)
end

def change_option_nocurves
	@option_nocurves = !@option_nocurves
end

def change_option_cpoint
	@option_cpoint = !@option_cpoint
end

def change_option_diameter
	@option_diameter = !@option_diameter
	onSetCursor
end
	
	
def onLButtonDown(flags, x, y, view)
	@time_mouse_down = Time.now
	@xdown = x
    @ydown = y
	if (@state == STATE_END)
		return if (@xdown - @xorig).abs < 2 && (@ydown - @yorig).abs < 2
	end	
	set_state @state + 1
end

#Key Up
def onKeyUp(key, rpt, flags, view)
	key = Traductor.check_key key, flags, true

	case key
		#Toggling between fixed and variable length
		when COPY_MODIFIER_KEY
			change_option_linemode if @control_down
			onMouseMove(flags, @xend, @yend, view) if (@state >= STATE_END)
			view.invalidate
			info_show
	end	
	@control_down = false
end

#Key down
def onKeyDown(key, rpt, flags, view)
	key = Traductor.check_key key, flags, false

	case key
		#Toggling between line and Cline mode
		when COPY_MODIFIER_KEY
			@control_down = true
			return
			
		#Calling options
		when TABLE_FKEY[OFS___Group]
			change_option_group
		when TABLE_FKEY[OFS___CPoint]
			change_option_cpoint
		when TABLE_FKEY[OFS___NoCurves]
			change_option_nocurves
		when TABLE_FKEY[OFS___Diameter]
			change_option_diameter
		when TABLE_FKEY[OFS___LineMode]
			change_option_linemode
			
		when 13
			@enter_down = true
	end	
	@control_down = false
	
	onMouseMove(flags, @xend, @yend, view) if (@state >= STATE_END)
	view.invalidate
	info_show
end

#Double Click to repeat with same length
def onLButtonDoubleClick(flags, x, y, view)
	redo_previous
end

def push_previous(pts)
	nb = pts.length - 2
	d = 0
	if (nb >= 0)
		d = 0
		for i in 0..nb
			d += pts[i].distance pts[i+1]
		end	
	end	
	return if d == 0
	
	#swapping the stored values	
	@prev_dist = @prev_dist0
	@prev_vec = (@prev_vec0) ? @prev_vec0.clone : nil
	@prev_dist0 = d
	@prev_vec0 = pts[0].vector_to pts[1]
end

def redo_previous
	if (@pt_end && @prev_dist > 0 && @moved)
		#correct_previous @prev_dist
	elsif (@moved == false && @prev_vec0 && @prev_dist0 > 0)
		mark = @mark_origin
		execute_to_distance(mark, @prev_vec0, @prev_dist0)
	else
		UI.beep
	end	
end

def correct_previous(len)
	#Checking if this is applicable
	return false unless @edges
	@edges.each { |e| return false unless e.valid? }
	return false unless @parcours_radius && @parcours_radius.length >= 2
	
	#Computing the origin
	mark = @parcours_radius[0]
	vecdir = mark.pt.vector_to @parcours_radius[1].pt
		
	Sketchup.undo
	execute_to_distance mark, vecdir, len
	true
end

#Redo operation when called from contextual menu
def menu_redo
	if (@pt_end && @prev_dist0 > 0 && @moved)
		vecdir = @pts_radius[0].vector_to @pts_radius[1]
		execute_to_distance(@mark_origin, vecdir, @prev_dist0)
	elsif (@moved == false && @prev_vec0 && @prev_dist0 > 0)
		execute_to_distance(@mark_origin, @prev_vec0, @prev_dist0)
	else
		UI.beep
	end	
end

#entry in the VCB: Distance and nb of segments
def onUserText(text, view) 
	@enter_down = false
	lstring = text.split ";"
	lseg = []
	ldist = []
	lstring.each do |s|
		str = s.strip
		next if str.length == 0
		if (str =~ /s$/i)
			lseg.push $`
		else
			ldist.push str
		end	
	end
	text_seg = (lseg.length > 0) ? lseg.last : nil
	text_dist = (ldist.length > 0) ? ldist.last : nil
	change_from_VCB view, text_dist, text_seg
end

#Change both distance and number of segments from the VCB
def change_from_VCB(view, text_dist, text_seg) 

	#forcing distance
	if text_dist
		begin
			len = text_dist.to_l
			if len == 0
				UI.beep
				return				
			end	
		rescue
			UI.beep
			return
		end
	else
		len = @prev_dist
	end	
		
	
	#Changing number of segments
	if text_seg
		begin
			nbs = text_seg.to_i
			if nbs <= 2 || nbs > 150
				UI.beep
				return
			elsif nbs == @nb_seg
				return
			end	
			@nb_seg = nbs
			unless text_dist
				view.invalidate
				info_show
				return
			end	
		rescue
			UI.beep
		end
	end	
	
	#Modify the distance for the previously entered line
	if @state == STATE_ORIGIN || @pt_end == nil || @pt_end == @pt_origin || @distance == 0
		UI.beep unless correct_previous len
		return
	elsif @pt_end
		vecdir = @pt_origin.vector_to @pt_end
		if vecdir.length > 0.0
			mark = OFSG.create_mark @pt_origin, @face_origin
			execute_to_distance mark, vecdir, len
		else
			UI.beep
		end	
	end	
end

#Mouse Move method
def onMouseMove(flags, x, y, view)
	view.lock_inference
	#Origin Point
    if (@state == STATE_ORIGIN)
		@xorig = x
		@yorig = y
        @ip.pick view, x, y
        
		@ip_origin.copy! @ip
		@mark_origin = OFSG.mark_from_inputpoint view, @ip_origin, x, y
		@pt_origin = @mark_origin.pt
		@face_origin = @mark_origin.face
        
		@pt_end = nil
		@moved = false
		
	#End Point	
	elsif (@state == STATE_END)
		@xend = x
		@yend = y		
        @ip.pick view, x, y
        if (@ip != @ip_origin) && @ip.valid?
            @ip_end.copy! @ip
			@mark_end = OFSG.mark_from_inputpoint view, @ip_end, x, y
			@pt_end = @mark_end.pt
			@face_end = @mark_end.face
			@moved = true
        end
	
	end
	tt = @ip.tooltip + "  [#{@nb_seg} #{@msg_Segment}]"
	view.tooltip = tt
	view.invalidate
	info_show
end	

#Draw method for tool
def draw(view)
	if (@state >= STATE_ORIGIN)
	    if @ip_origin.valid? && @ip_origin.display?
			if (@face_origin == nil)
				view.draw_points @pt_origin, 15, 3, 'purple' if @pt_origin
			elsif @pt_origin && ! OFSG.within_face_extended?(@face_origin, @pt_origin)
				view.draw_points @pt_origin, 15, 5, 'orange'
			else	
				@ip_origin.draw view
				view.draw_points @pt_origin, 15, 3, 'purple' if @pt_origin && @pt_origin != @ip_origin.position
			end	
       end
	end	
	
	if (@state >= STATE_END)
	    if @ip_end.valid? && @ip_end.display? && @pt_end
			if (@face_end == nil)
				view.draw_points @pt_end, 15, 3, 'purple'
			elsif @face_end.classify_point(@pt_end) >= 8 
				view.draw_points @pt_end, 15, 5, 'orange'
			else	
				@ip_end.draw view
			end	
        end
		
		compute_path_radius
		pts = @pts_radius
		view.drawing_color = 'red'
		view.line_stipple = "_"
		view.line_width = 2
		view.draw GL_LINE_STRIP, pts if pts.length > 1
		if (@option_group)
			color = OFS_COLOR_Group
			width = 4
			cwidth = 2
		else
			color = OFS_COLOR_Normal
			width = 2
			cwidth = 1
		end	
		if (@linemode)
			view.drawing_color = color
			view.line_width = width
			view.line_stipple = (@option_group) ? "-.-" : ""
		else
			view.drawing_color = color
			view.line_width = cwidth
			view.line_stipple = "_"
		end
		view.draw GL_LINE_STRIP, @pts_contour if @pts_contour.length > 1
		view.line_stipple = ""
		view.draw_points @pts_vertices, 10, 5, color if @pts_vertices && @pts_vertices.length > 2
		view.draw_points pts[1..-2], 10, 3, 'black' if @option_cpoint && pts.length > 2
	end	
	#OFSG.watchlist_draw view
end

#display information in the Sketchup status bar
def info_show
	msg = "[" + Traductor[(@linemode) ? STR_ModeLine : STR_ModeCLine] + "]"
	msg += " #{@nb_seg} #{@msg_Segment} -- "

	case @state
	when STATE_ORIGIN
		msg += @msg_Polygon_Origin
	when STATE_END
		msg += @msg_Polygon_End
	when STATE_EXECUTION
		msg += @msg_Polygon_Execution
	end
	
	d = @distance

	msg += " [" + @msg_Group + "]" if (@option_group)
	msg += " [" + @msg_CPoint + "]" if (@option_cpoint)
	msg += " [" + @msg_Diameter + "]" if (@option_diameter)
	
	Sketchup.set_status_text msg	
	label = (@option_diameter) ? @msg_Diameter : @msg_Radius
	Sketchup.set_status_text label, SB_VCB_LABEL
	Sketchup.set_status_text d.to_l, SB_VCB_VALUE
end

end	#Class TOSToolPolygon
	
end	#End Module SUToolsOnSurface
