=begin
#-------------------------------------------------------------------------------------------------------------------------------------------------
#*************************************************************************************************
# Designed April / June 2008 by Fredo6

# Permission to use, copy, modify, and distribute this software for 
# any purpose and without fee is hereby granted, provided that the above
# copyright notice appear in all copies.

# THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
#-----------------------------------------------------------------------------
# Name			:   Zloader__OnSurface.rb
# Original Date	:   04 June 2008 - version 1.2
# Type			:   Sketchup Tools
# Description	:   Loader for the Suite of Tools to draw on a surface
# Menu Items	:   Tools --> submenu "Tools on Surface"
# Toolbar		:   Name = "Surface Operations"
# Usage			:   See Tutorial and Quick Ref Card in PDF format
#-------------------------------------------------------------------------------------------------------------------------------------------------
#*************************************************************************************************
=end

require 'sketchup.rb'
require 'LibTraductor.rb'			# for language translation
require 'LineOnSurface.rb'
require 'PolygonOnSurface.rb'
require 'OffsetOnSurface.rb'
require 'EraserOnSurface.rb'
require 'PolylineOnSurface.rb'

module SUToolsOnSurface


#--------------------------------------------------------------------------------------------------------------
# Public methods of module for EraserOnSurface (Menu and Tool bar)
#--------------------------------------------------------------------------------------------------------------

def SUToolsOnSurface.manage_command
	menutool = SUToolsOnSurface.get_submenu
	@tlb = nil

	text = Traductor[STR_Line_Title]
	cmd = UI::Command.new(text) { SUToolsOnSurface.launch_line }
	SUToolsOnSurface.add_command(cmd, menutool, text, text, DOS_ICON_LINE)
	
	text = Traductor[STR_Circle_Title]
	cmd = UI::Command.new(text) { SUToolsOnSurface.launch_circle }
	SUToolsOnSurface.add_command(cmd, menutool, text, text, DOS_ICON_CIRCLE_LINE)
	
	text = Traductor[STR_Polygon_Title]
	cmd = UI::Command.new(text) { SUToolsOnSurface.launch_polygon }
	SUToolsOnSurface.add_command(cmd, menutool, text, text, DOS_ICON_POLYGON_LINE)

	title = Traductor[STR_Ofs_Title]
	tooltip = Traductor[STR_Ofs_Tooltip]
	cmd = UI::Command.new(title) { SUToolsOnSurface.OFS_launch true }	
	SUToolsOnSurface.add_command(cmd, menutool, title, tooltip, OFS_ICON_OFFSET)

	@tlb.add_separator if @tlb
	
	text = Traductor[STR_PolyLine_Title]
	cmd = UI::Command.new(text) { SUToolsOnSurface.launch_polyline }
	SUToolsOnSurface.add_command(cmd, menutool, text, text, DOS_ICON_POLYLINE)
	
	text = Traductor[STR_Eraser_Title]
	cmd = UI::Command.new(text) { SUToolsOnSurface.launch_eraser }
	SUToolsOnSurface.add_command(cmd, menutool, text, text, DOS_ICON_ERASER)
	
	@tlb.show if @tlb

	UI.add_context_menu_handler do |menu|
		ledge = SUToolsOnSurface.check_if_polyline Sketchup.active_model.selection
		if ledge
			menu.add_separator
			menu.add_item(Traductor[STR_PolyLine_Title]) { SUToolsOnSurface.launch_polyline ledge}
		end	
		if VoidTriangle.selection_check_for_menu(false)
			menu.add_item(Traductor[STR_VoidTriangle]) { VoidTriangle.selection_check_for_menu true}
		end	
	end
	
end

unless $Loader__OnSurface____loaded
	SUToolsOnSurface.manage_command
	$Loader__OnSurface____loaded = true
end
	
end	#End Module SUToolsOnSurface
